import typing

import pydantic as pydantic_
from pydantic_core import CoreSchema, core_schema, SchemaValidator
from pydantic_core.core_schema import SerializerFunctionWrapHandler, ValidationInfo

from typenum.core import NoValue, TypEnumContent
from typenum.pydantic.serialization.tagged import TaggedSerialization

if typing.TYPE_CHECKING:
    from ..core import TypEnumPydantic  # type: ignore

__all__ = [
    "InternallyTagged",
]


class InternallyTagged(TaggedSerialization):
    __variant_tag__: str
    __ext_tagged_schema_validator__: SchemaValidator

    def __init__(self, variant: str):
        self.__variant_tag__ = variant

    def __get_pydantic_core_schema__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            source_type: typing.Any,
            handler: pydantic_.GetCoreSchemaHandler,
    ) -> CoreSchema:
        from typenum.pydantic.core import TypEnumPydantic

        json_schemas: dict[str, core_schema.CoreSchema] = {}
        real_schema_attrs = {}
        real_schemas: list[core_schema.CoreSchema] = []

        for attr in kls.__variants__.values():
            enum_variant: type[TypEnumPydantic[TypEnumContent]] = getattr(kls, attr)
            attr = kls.__names_serialization__.get(attr, attr)
            variant_schema = core_schema.typed_dict_field(core_schema.str_schema(pattern=attr))

            schema = {
                self.__variant_tag__: variant_schema,
            }
            if enum_variant.__content_type__ is NoValue:
                real_schemas.append(core_schema.str_schema(pattern=attr))
            else:
                item_schema = handler.generate_schema(enum_variant.__content_type__)
                resolved = handler.resolve_ref_schema(item_schema)

                real_schema_attrs[attr] = core_schema.typed_dict_field(resolved, required=False)
                real_schemas.append(resolved)

                match resolved:
                    case {"type": "dataclass", "schema": {"fields": fields}}:
                        fields = {
                            field["name"]: {
                                "type": "typed-dict-field",
                                "schema": field["schema"]
                            } for field in fields
                        }
                        schema.update(**fields)
                    case {"type": "model", "schema": {"fields": fields}}:
                        fields = {
                            k: {
                                "type": "typed-dict-field",
                                "schema": v["schema"]
                            } for k, v in fields.items()
                        }
                        schema.update(**fields)
                    case {"type": "typed-dict", "fields": fields}:
                        schema.update(**fields)
                    case _:
                        raise TypeError(
                            "Type of content must be a TypedDict, dataclass or BaseModel subclass"
                        )

            json_schemas[attr] = core_schema.typed_dict_schema(schema)

        # Store nested schema for deserializing
        self.__ext_tagged_schema_validator__ = SchemaValidator(core_schema.union_schema([
            core_schema.typed_dict_schema(real_schema_attrs),
            *real_schemas,
        ]))

        json_schema = core_schema.tagged_union_schema(
            choices=json_schemas,
            discriminator=self.__variant_tag__,
        )
        return core_schema.json_or_python_schema(
            json_schema=core_schema.with_info_after_validator_function(
                kls.__python_value_restore__,
                json_schema,
            ),
            python_schema=core_schema.with_info_after_validator_function(
                kls.__python_value_restore__,
                core_schema.any_schema(),
            ),
            serialization=core_schema.wrap_serializer_function_ser_schema(
                kls.__pydantic_serialization__
            ),
            ref=f"{kls.__name__}:{id(kls)}"
        )

    def __python_value_restore__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            input_value: typing.Any,
            info: ValidationInfo,
    ) -> typing.Any:
        if isinstance(input_value, kls):
            return input_value

        type_key = input_value.pop(self.__variant_tag__)
        attr = kls.__names_deserialization__.get(type_key, type_key)

        if input_value:
            value = self.__ext_tagged_schema_validator__.validate_python({type_key: input_value})
            return getattr(kls, attr).__variant_constructor__(value[type_key], info)
        else:
            return getattr(kls, attr).__variant_constructor__(None, info)

    def __pydantic_serialization__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            model: typing.Any,
            serializer: SerializerFunctionWrapHandler,
    ) -> typing.Any:
        attr = model.__variant_name__
        attr = kls.__names_serialization__.get(attr, attr)

        result = {self.__variant_tag__: attr}
        if model.__content_type__ is NoValue:
            pass
        elif isinstance(model.value, kls):
            result.update(**kls.__pydantic_serialization__(model.value, serializer))
        else:
            result.update(**serializer(model.value))

        return result



import typing
from abc import ABC, abstractmethod

import pydantic as pydantic_
from pydantic_core import CoreSchema
from pydantic_core.core_schema import SerializerFunctionWrapHandler, ValidationInfo

if typing.TYPE_CHECKING:
    from ...core import TypEnumContent  # type: ignore
    from ..core import TypEnumPydantic  # type: ignore

__all__ = [
    "TaggedSerialization",
]


class TaggedSerialization(ABC):
    @abstractmethod
    def __get_pydantic_core_schema__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            _source_type: typing.Any,
            handler: pydantic_.GetCoreSchemaHandler,
    ) -> CoreSchema:
        raise NotImplementedError

    @abstractmethod
    def __python_value_restore__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            input_value: typing.Any,
            info: ValidationInfo,
    ) -> typing.Any:
        raise NotImplementedError

    @abstractmethod
    def __pydantic_serialization__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            model: typing.Any,
            serializer: SerializerFunctionWrapHandler,
    ) -> typing.Any:
        raise NotImplementedError



import inspect
import typing

import pydantic as pydantic_
from pydantic_core import CoreSchema, core_schema
from pydantic_core.core_schema import SerializerFunctionWrapHandler, ValidationInfo

from typenum.core import TypEnumContent, NoValue
from typenum.pydantic.serialization.tagged import TaggedSerialization

if typing.TYPE_CHECKING:
    from ..core import TypEnumPydantic  # type: ignore


__all__ = [
    "AdjacentlyTagged",
]


class AdjacentlyTagged(TaggedSerialization):
    __variant_tag__: str
    __content_tag__: str

    def __init__(self, variant: str, content: str):
        self.__variant_tag__ = variant
        self.__content_tag__ = content

    def __get_pydantic_core_schema__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            _source_type: typing.Any,
            handler: pydantic_.GetCoreSchemaHandler,
    ) -> CoreSchema:
        from typenum.pydantic.core import TypEnumPydantic

        json_schemas: list[core_schema.CoreSchema] = []
        for attr in kls.__variants__.values():
            enum_variant: type[TypEnumPydantic[TypEnumContent]] = getattr(kls, attr)
            attr = kls.__names_serialization__.get(attr, attr)
            variant_schema = core_schema.typed_dict_field(core_schema.str_schema(pattern=attr))
            is_typenum_variant = (
                    inspect.isclass(enum_variant.__content_type__) and
                    issubclass(enum_variant.__content_type__, TypEnumPydantic)
            )

            schema = {
                self.__variant_tag__: variant_schema,
            }

            if is_typenum_variant or enum_variant.__content_type__ is NoValue:
                if is_typenum_variant:
                    kls_: type = enum_variant.__content_type__  # type: ignore
                    schema_definition = core_schema.definition_reference_schema(f"{kls_.__name__}:{id(kls_)}")
                    value_schema = core_schema.typed_dict_field(core_schema.definitions_schema(
                        schema=schema_definition,
                        definitions=[
                            core_schema.any_schema(ref=f"{kls_.__name__}:{id(kls_)}")
                        ],
                    ))

                    schema[self.__content_tag__] = value_schema
            else:
                value_schema = core_schema.typed_dict_field(handler.generate_schema(enum_variant.__content_type__))
                schema[self.__content_tag__] = value_schema

            json_schemas.append(core_schema.typed_dict_schema(schema))

        return core_schema.json_or_python_schema(
            json_schema=core_schema.with_info_after_validator_function(
                kls.__python_value_restore__,
                core_schema.union_schema([*json_schemas]),
            ),
            python_schema=core_schema.with_info_after_validator_function(
                kls.__python_value_restore__,
                core_schema.union_schema([*json_schemas, core_schema.any_schema()]),
            ),
            serialization=core_schema.wrap_serializer_function_ser_schema(
                kls.__pydantic_serialization__
            ),
            ref=f"{kls.__name__}:{id(kls)}"
        )

    def __python_value_restore__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            input_value: typing.Any,
            info: ValidationInfo,
    ) -> typing.Any:
        from typenum.pydantic.core import TypEnumPydantic

        if isinstance(input_value, TypEnumPydantic):
            return input_value

        type_key = input_value[self.__variant_tag__]
        value = input_value.get(self.__content_tag__, None)
        attr = kls.__names_deserialization__.get(type_key, type_key)
        return getattr(kls, attr).__variant_constructor__(value, info)

    def __pydantic_serialization__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            model: typing.Any,
            serializer: SerializerFunctionWrapHandler,
    ) -> typing.Any:
        attr = model.__variant_name__
        attr = kls.__names_serialization__.get(attr, attr)

        result = {self.__variant_tag__: attr}
        if model.__content_type__ is NoValue:
            pass
        elif isinstance(model.value, kls):
            result[self.__content_tag__] = kls.__pydantic_serialization__(model.value, serializer)
        else:
            result[self.__content_tag__] = serializer(model.value)

        return result



import types
import typing

__all__ = [
    "NoValue",
    "TypEnum",
    "TypEnumContent",
    "TypEnumMeta",
]

import typing_extensions

from annotated_types import BaseMetadata
from typing_extensions import Annotated

TypEnumContent = typing.TypeVar("TypEnumContent")


NoValue = types.EllipsisType


class TypEnumMeta(type):
    __full_variant_name__: str
    __variant_name__: str

    __content_type__: typing.Union[str, type[typing.Any]]

    __variants__: dict[type['_TypEnum[typing.Any]'], str]

    __is_variant__: bool = False

    def __new__(
            cls,
            cls_name: str,
            bases: tuple[typing.Any],
            class_dict: dict[str, typing.Any],
    ) -> typing.Any:
        enum_class = super().__new__(cls, cls_name, bases, class_dict)
        if enum_class.__annotations__.get("__abstract__"):
            return enum_class

        if enum_class.__is_variant__:
            return enum_class
        else:
            enum_class.__variants__ = dict()

        enum_class.__full_variant_name__ = cls_name
        enum_class.__variant_name__ = cls_name

        annotation: typing.Union[type[Annotated[typing.Any, BaseMetadata]], type]
        for attr, annotation in enum_class.__annotations__.items():
            if not hasattr(annotation, "__args__"):
                continue

            if (__origin__ := getattr(annotation, "__origin__", None)) and annotation.__name__ == "Annotated":
                origin = typing.get_args(__origin__)[0]
            else:
                is_type = isinstance(annotation, types.GenericAlias) and annotation.__name__ == "type"
                if not is_type:
                    continue

                origin = typing.get_args(annotation)[0]

            split = origin[:-1].split("[", maxsplit=1)

            content_type: str | type[typing.Any]
            if len(split) == 1:
                content_type = NoValue
            else:
                left, right = split
                if left != enum_class.__name__:
                    continue

                if right.split("[", maxsplit=1)[0] == enum_class.__name__:
                    content_type = enum_class
                else:
                    try:
                        content_type = eval(right)
                    except NameError:
                        content_type = right

            try:
                variant_base = enum_class[content_type]  # type: ignore
            except TypeError:
                # When enum is non-generic, like this
                #
                # class SimpleEnum(TypEnum):
                #     V: type["SimpleEnum"]
                #
                variant_base = enum_class

            class _EnumVariant(variant_base):  # type: ignore
                __is_variant__ = True

            _EnumVariant.__name__ = _EnumVariant.__full_variant_name__ = f"{enum_class.__name__}.{attr}"
            _EnumVariant.__variant_name__ = attr
            _EnumVariant.__content_type__ = content_type

            enum_class.__variants__[_EnumVariant] = attr

            setattr(enum_class, attr, _EnumVariant)

        return enum_class

    def __repr__(self) -> str:
        return getattr(self, "__full_variant_name__", self.__class__.__name__)


class _TypEnum(typing.Generic[TypEnumContent], metaclass=TypEnumMeta):
    __match_args__ = ("value",)

    __full_variant_name__: typing.ClassVar[str]
    __variant_name__: typing.ClassVar[str]

    __content_type__: typing.ClassVar[typing.Union[str, type[typing.Any]]]

    __variants__: typing.ClassVar[dict[type['_TypEnum[typing.Any]'], str]]

    __is_variant__: typing.ClassVar[bool] = False

    __abstract__: typing_extensions.Never

    value: typing.Optional[TypEnumContent]

    def __init__(self, value: TypEnumContent):
        if self.__content_type__ is NoValue:
            self.value = None
        else:
            self.value = value

    def __repr__(self) -> str:
        if self.__content_type__ is NoValue:
            return f"{self.__full_variant_name__}()"
        return f"{self.__full_variant_name__}({self.value.__repr__()})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _TypEnum):
            return False

        return self.__class__ == other.__class__ and self.value == other.value


class TypEnum(_TypEnum[TypEnumContent]):
    __abstract__: typing_extensions.Never
